# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import torch


def leaky_hardtanh(
    input: torch.Tensor,
    min_val: float = -1.0,
    max_val: float = 1.0,
    negative_slope: float = 0.01,
    positive_slope: float = 0.01,
) -> torch.Tensor:
    """Leaky version of hardtanh where regions outside [min_val, max_val] have small non-zero slopes.

    Args:
        input: Input tensor
        min_val: Minimum value for the hardtanh region
        max_val: Maximum value for the hardtanh region
        negative_slope: Slope for values below min_val
        positive_slope: Slope for values above max_val

    Returns:
        Tensor with leaky hardtanh applied
    """
    below_min = input < min_val
    above_max = input > max_val
    # Standard hardtanh behavior for the middle region
    result = torch.clamp(input, min_val, max_val)
    # Add leaky behavior for regions outside the clamped range
    result = torch.where(below_min, min_val + negative_slope * (input - min_val), result)
    result = torch.where(above_max, max_val + positive_slope * (input - max_val), result)
    return result



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from typing import Optional

from anemoi.models.data_indices.collection import IndexCollection
from anemoi.models.preprocessing import BasePreprocessor
from anemoi.models.preprocessing.monomapper import Monomapper
from anemoi.models.preprocessing.multimapper import Multimapper

LOGGER = logging.getLogger(__name__)


class Remapper(BasePreprocessor, ABC):
    """Remap and convert variables for single variables."""

    def __new__(
        cls,
        config=None,
        data_indices: Optional[IndexCollection] = None,
        statistics: Optional[dict] = None,
    ) -> None:
        _, _, method_config = cls._process_config(config)
        monomappings = Monomapper.supported_methods
        multimappings = Multimapper.supported_methods
        if all(method in monomappings for method in method_config):
            return Monomapper(config, data_indices, statistics)
        elif all(method in multimappings for method in method_config):
            return Multimapper(config, data_indices, statistics)
        elif not (
            any(method in monomappings for method in method_config)
            or any(method in multimappings for method in method_config)
        ):
            raise ValueError("No valid remapping method found.")
        else:
            raise NotImplementedError(
                f"Not implemented: method_config contains a mix of monomapper and multimapper methods: {list(method_config.keys())}"
            )



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from enum import Enum

from pydantic import Field
from pydantic import NonNegativeInt

from anemoi.utils.schemas import BaseModel


class ActivationFunctons(str, Enum):
    GELU = "GELU"
    SiLU = "SiLU"
    ELU = "ELU"
    ReLU = "ReLU"
    Tanh = "Tanh"
    Sigmoid = "Sigmoid"
    Hardshrink = "Hardshrink"
    Hardsigmoid = "Hardsigmoid"
    Hardtanh = "Hardtanh"
    Hardswish = "Hardswish"
    LeakyReLU = "LeakyReLU"
    LogSigmoid = "LogSigmoid"
    PReLU = "PReLU"
    ReLU6 = "ReLU6"
    SELU = "SELU"
    CELU = "CELU"
    Mish = "Mish"
    Softplus = "Softplus"
    Softshrink = "Softshrink"
    Softsign = "Softsign"
    Tanhshrink = "Tanhshrink"
    Threshold = "Threshold"
    GLU = "GLU"


class TransformerModelComponent(BaseModel):
    activation: ActivationFunctons = Field(example="GELU")
    "Activation function to use for the transformer model component. Default to GELU."
    convert_: str = Field("all", alias="_convert_")
    "Target's parameters to convert to primitive containers. Other parameters will use OmegaConf. Default to all."
    cpu_offload: bool = Field(example=False)
    "Offload to CPU. Default to False."
    num_chunks: NonNegativeInt = Field(example=1)
    "Number of chunks to divide the layer into. Default to 1."
    mlp_hidden_ratio: NonNegativeInt = Field(example=4)
    "Ratio of mlp hidden dimension to embedding dimension. Default to 4."
    num_heads: NonNegativeInt = Field(example=16)
    "Number of attention heads. Default to 16."


class GNNModelComponent(BaseModel):
    activation: ActivationFunctons = Field(example="GELU")
    "Activation function to use for the GNN model component. Default to GELU."
    convert_: str = Field("all", alias="_convert_")
    "Target's parameters to convert to primitive containers. Other parameters will use OmegaConf. Default to all."
    trainable_size: NonNegativeInt = Field(example=8)
    "Size of trainable parameters vector. Default to 8."
    num_chunks: NonNegativeInt = Field(example=1)
    "Number of chunks to divide the layer into. Default to 1."
    cpu_offload: bool = Field(example=False)
    "Offload to CPU. Default to False."
    sub_graph_edge_attributes: list[str] = Field(default_factory=list)
    "Edge attributes to consider in the model component features."
    mlp_extra_layers: NonNegativeInt = Field(example=0)
    "The number of extra hidden layers in MLP. Default to 0."



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from hydra.errors import InstantiationException
from hydra.utils import instantiate
from torch import nn
from torch.utils.checkpoint import checkpoint

from anemoi.utils.config import DotDict

LOGGER = logging.getLogger(__name__)


class CheckpointWrapper(nn.Module):
    """Wrapper for checkpointing a module."""

    def __init__(self, module: nn.Module) -> None:
        super().__init__()
        self.module = module

    def forward(self, *args, **kwargs):
        return checkpoint(self.module, *args, **kwargs, use_reentrant=False)


def load_layer_kernels(kernel_config: DotDict) -> DotDict:
    """Load layer kernels from the config.

    Args:
        kernel_config : DotDict
            Kernel configuration, e.g. {"Linear": {"_target_": "torch.nn.Linear", "_partial_": True}}

    Returns:
        DotDict: hydra partial instantiation of the layer kernels
    """
    # If self.layer_kernels entry is missing from the config, use torch.nn kernels
    default_kernels = {
        "Linear": {"_target_": "torch.nn.Linear", "_partial_": True},
        "LayerNorm": {"_target_": "torch.nn.LayerNorm", "_partial_": True},
        "QueryNorm": {
            "_target_": "anemoi.models.layers.normalization.AutocastLayerNorm",
            "_partial_": True,
            "bias": False,
        },
        "KeyNorm": {
            "_target_": "anemoi.models.layers.normalization.AutocastLayerNorm",
            "_partial_": True,
            "bias": False,
        },
    }
    layer_kernels = {**default_kernels, **kernel_config}

    # Loop through all kernels in the layer_kernels config entry and try import them
    for kernel in layer_kernels:
        kernel_entry = layer_kernels[kernel]
        try:
            instantiate(kernel_entry)
        except InstantiationException:
            LOGGER.info(
                f"{kernel_entry['_target_']} not found! check your config.model.layer_kernel.{kernel} entry. Maybe your desired kernel is not installed or the import string is incorrect?"
            )
            raise InstantiationException
        else:
            LOGGER.info(f"{kernel} kernel: {kernel_entry}")
    return layer_kernels



# (C) Copyright 2024- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from typing import Literal

from pydantic import Field
from pydantic import NonNegativeInt

from .common_components import GNNModelComponent
from .common_components import TransformerModelComponent


class GNNDecoderSchema(GNNModelComponent):
    target_: Literal["anemoi.models.layers.mapper.GNNBackwardMapper"] = Field(..., alias="_target_")
    "GNN decoder object from anemoi.models.layers.mapper."


class GraphTransformerDecoderSchema(TransformerModelComponent):
    target_: Literal["anemoi.models.layers.mapper.GraphTransformerBackwardMapper"] = Field(..., alias="_target_")
    "Graph Transformer Decoder object from anemoi.models.layers.mapper."
    trainable_size: NonNegativeInt = Field(example=8)
    "Size of trainable parameters vector. Default to 8."
    sub_graph_edge_attributes: list[str] = Field(example=["edge_length", "edge_dirs"])
    "Edge attributes to consider in the decoder features. Default to [edge_length, edge_dirs]"
    qk_norm: bool = Field(example=False)
    "Normalize the query and key vectors. Default to False."
    initialise_data_extractor_zero: bool = Field(example=False)
    "Initialise the data extractor with zeros. Default to False."



import pytest
import torch
from hydra.errors import InstantiationException
from hydra.utils import instantiate
from omegaconf import OmegaConf

from anemoi.models.layers.utils import load_layer_kernels


@pytest.fixture
def default_layer_kernels():
    # Default layer kernels
    kernels_config = OmegaConf.create(
        {
            "LayerNorm": {
                "_target_": "torch.nn.LayerNorm",
                "_partial_": True,
            },
        }
    )
    return instantiate(load_layer_kernels(kernels_config))


@pytest.fixture
def custom_layer_kernels():
    # Custom layer kernels
    kernels_config = OmegaConf.create(
        {
            "LayerNorm": {
                "_target_": "torch.nn.LayerNorm",
                "_partial_": True,
                "eps": 1e-3,
                "elementwise_affine": False,
            },
            "Linear": {"_target_": "torch.nn.Linear", "_partial_": True, "bias": False},
        }
    )
    return instantiate(load_layer_kernels(kernels_config))


def test_kernels_init(default_layer_kernels):
    """Test that the layer kernels are instantiated."""
    channels = 10
    linear_layer = default_layer_kernels["Linear"](in_features=channels, out_features=channels)
    layer_norm = default_layer_kernels["LayerNorm"](normalized_shape=channels)
    assert isinstance(linear_layer, torch.nn.Linear)
    assert isinstance(layer_norm, torch.nn.LayerNorm)
    assert linear_layer.bias.shape == torch.Size([channels])
    assert layer_norm.bias.shape == torch.Size([channels])


def test_custom_kernels(custom_layer_kernels):
    """Test that the custom layer kernels are instantiated."""
    linear_layer = custom_layer_kernels["Linear"](in_features=10, out_features=10)
    layer_norm = custom_layer_kernels["LayerNorm"](normalized_shape=10)

    assert linear_layer.bias is None
    assert layer_norm.bias is None


def test_unavailable_kernel():
    """Config with an unavailable kernel that should raise an error."""
    kernels_config = OmegaConf.create(
        {
            "LayerNorm": {"_target_": "nonexistent_package.LayerNorm", "_partial_": True},
            "Linear": {"_target_": "torch.nn.Linear", "_partial_": True},
        }
    )
    # Catch InstantiationException
    with pytest.raises(InstantiationException):
        load_layer_kernels(kernels_config)
